<?php

/**
 * @file
 * Provide class that manages memcache connections.
 */

/**
 * Integrates with memcache API.
 */
class MemcacheStorageAPI {

  // List of available memcache connections.
  private static $connections;

  // Static counter of timer.
  // Requires to display time of execution for every memcache operation.
  private static $debug_timer = 0;

  /**
   * Connect to memcache clustes and returns memcache object on success.
   *
   * @param $bin
   *   Cache bin name. i.e. 'cache' or 'cache_bootstrap'.
   *
   * @return bool|object
   *   Memcache or memcached object is connection was successfull. Otherwise FALSE.
   */
  public static function openConnection($bin = '') {

    // Get name of cluster for selected cache bin storage.
    $bin_cluster = self::getCacheBinCluster($bin);

    if (empty(self::$connections[$bin_cluster])) {

      // Load available servers and its clusters from settings.php.
      $server_list = variable_get('memcache_servers', array('127.0.0.1:11211' => 'default'));

      // Search for all servers matching bin cluster.
      $cluster_servers = array();
      foreach ($server_list as $server => $cluster_name) {
        if ($cluster_name == $bin_cluster) {
          $cluster_servers[] = $server;
        }
      }

      // No servers found for this cache bin.
      if (empty($cluster_servers)) {

        // Log error.
        self::logError('Could not find server(s) for cluster %cluster (bin %bin). Check your settings.php configuration.',
          array('%cluster' => $bin_cluster, '%bin' => $bin));

        // Return not initialized value.
        return FALSE;
      }

      $preferred = variable_get('memcache_extension');
      if (isset($preferred) && class_exists($preferred)) {
        $extension = $preferred;
      }
      // If no extension is set, default to PECL Memcache.
      elseif (class_exists('Memcache')) {
        $extension = 'Memcache';
      }
      elseif (class_exists('Memcached')) {
        $extension = 'Memcached';
      }

      // Check if Memcache extension enabled.
      if (empty($extension)) {

        // Log error.
        self::logError('<a href="@url_memcache">Memcache</a> or <a href="@url_memcached">Memcached</a> extension is not available.',
          array('@url_memcache' => 'http://pecl.php.net/package/memcache', '@url_memcached' => 'http://pecl.php.net/package/memcache'));

        return FALSE;
      }

      // Create new memcache connection.
      $memcache = new $extension();

      // Collect all servers thas is unable to connect.
      $failed_connections = array();

      // Connect to every server.
      foreach ($cluster_servers as $server) {

        list($host, $port) = explode(':', $server);

        // Support unix sockets in the format 'unix:///path/to/socket'.
        if ($host == 'unix') {

          // PECL Memcache supports 'unix:///path/to/socket' path in ::addServer function,
          // while PECL Memcached use only '/path/to/socket' string for the same purpose.
          if ($extension == 'Memcache') {
            $host = $server;
          }
          elseif ($extension == 'Memcached') {
            $host = substr($server, 7);
          }

          // Port is always 0 for unix sockets.
          $port = 0;
        }

        // Adding new server for memcache connection.
        $connected = @$memcache->addServer($host, $port);

        if (!$connected) {

          // Mark connection to this server as 'failed'.
          $failed_connections[] = $server;

          // Log error if unable to connect to server.
          self::logError('Could not connect to server %server (port %port).',
            array('%server' => $host, '%port' => $port));
        }
      }

      // If memcache is unable to connect to all servers it means that
      // we have no connections at all, and could not start memcached connection.
      if (sizeof($failed_connections) == sizeof($cluster_servers)) {
        return FALSE;
      }

      // Provided compressThreshold support for PECL Memcached library.
      if ($extension == 'Memcache') {

        // For more information see http://www.php.net/manual/en/memcache.setcompressthreshold.php.
        $compress_threshold = variable_get('memcache_storage_compress_threshold', array('threshold' => 20000, 'min_savings' => 0.2));
        if (isset($compress_threshold['threshold']) && isset($compress_threshold['min_savings'])) {
          $memcache->setCompressThreshold($compress_threshold['threshold'], $compress_threshold['min_savings']);
        }
      }
      elseif ($extension == 'Memcached') {

        $default_options = array(
          Memcached::OPT_COMPRESSION => FALSE,
          Memcached::OPT_DISTRIBUTION => Memcached::DISTRIBUTION_CONSISTENT,
        );

        // For more info about memcached contants see http://www.php.net/manual/en/memcached.constants.php.
        $memcached_options = variable_get('memcache_options', array());
        $memcached_options += $default_options;

        // Set memcached options.
        // See http://www.php.net/manual/en/memcached.setoption.php.
        foreach ($memcached_options as $key => $value) {
          $memcache->setOption($key, $value);
        }
      }

      // Store memcache object as a static object, to avoid duplicate connections to the same
      // memcache cluster, but for differens cache bins.
      self::$connections[$bin_cluster] = $memcache;
    }

    return self::$connections[$bin_cluster];
  }

  /**
   * Close all memcache connections.
   *
   * @param $bin
   *   Cache bin name. i.e. 'cache' or 'cache_bootstrap'.
   */
  static public function closeConnection($bin) {

    // Get name of cluster for current cache bin storage.
    $cluster = self::getCacheBinCluster($bin);

    // Close memcache or memcached connection and destroy object.
    if (!empty(self::$connections[$cluster])) {
      $object = self::$connections[$cluster];
      if ($object instanceof Memcache) {
        $object->close();
      }
      elseif ($object instanceof Memcached) {
        $object->quit();
      }
      unset(self::$connections[$cluster]);
    }
  }

  /**
   * Get values from memcache storage.
   * Provide debug logging.
   *
   * @see http://www.php.net/manual/en/memcache.get.php
   */
  public static function get($cache_ids, $memcache, $cache_bin = '') {

    // Check the debug mode.
    $debug = variable_get('memcache_storage_debug', FALSE);

    /* START DEBUG. FIRST PHASE. */
    if ($debug) {
      // Start count time spent on getting data from memcache.
      $timer = $cache_bin . '_' . self::$debug_timer;
      timer_start($timer);
      // Get memory usage before set new value into cache.
      $memory_usage_start = memory_get_usage();
    }
    /* STOP DEBUG. END OF FIRST PHASE. */

    // If cache ids is a string, convert it to array anyway.
    if (!is_array($cache_ids)) {
      $cache_ids = array($cache_ids);
    }

    // Get data from memcache storage.
    $cache = array();
    if ($memcache instanceof Memcache) {
      $cache = $memcache->get($cache_ids);
    }
    elseif ($memcache instanceof Memcached) {
      $cache = $memcache->getMulti($cache_ids);
    }

    /* START DEBUG. LAST PHASE. */
    if ($debug) {

      // Stop count timer.
      timer_stop($timer);

      // Calculate memory usage.
      $used_memory = memory_get_usage() - $memory_usage_start;

      // Get array with debug data.
      $memcache_storage_debug_output = &drupal_static('memcache_storage_debug_output');

      // Log entry about memcache action.
      foreach ($cache_ids as $cache_id) {
        $memcache_storage_debug_output[] = array(
          'action' => count($cache_ids) > 1 ? 'getMultiple' : 'get',
          'timer' => timer_read($timer),
          'memory' => number_format($used_memory / 1024, 2),
          'bin' => $cache_bin,
          'key' => urldecode($cache_id),
          'result' => isset($cache[$cache_id]) ? 'HIT' : 'MISS',
        );
      }

      // Increase timer to get new time results.
      self::$debug_timer++;
    }
    /* END DEBUG. LAST PHASE. */

    return $cache;
  }

  /**
   * Set data to memcache pool.
   * Provide debug logging.
   *
   * @see http://www.php.net/manual/en/memcache.set.php
   */
  public static function set($cache_id, $data, $expire, $memcache, $cache_bin = '') {

    // Check the debug mode.
    $debug = variable_get('memcache_storage_debug', FALSE);

    /* START DEBUG. FIRST PHASE. */
    if ($debug) {
      // Start count time spent on setting data to memcache.
      $timer = $cache_bin . '_' . self::$debug_timer;
      timer_start($timer);
      // Get memory usage before set new value into cache.
      $memory_usage_start = memory_get_usage();
    }
    /* STOP DEBUG. END OF FIRST PHASE. */

    // Add data to memcached pool.
    $result = FALSE;
    if ($memcache instanceof Memcache) {
      // Get compression mode.
      $compressed = variable_get('memcache_storage_compress_data', 0);
      $result = $memcache->set($cache_id, $data, $compressed, $expire);
    }
    elseif ($memcache instanceof Memcached) {
      $result = $memcache->set($cache_id, $data, $expire);
    }

    /* START DEBUG. LAST PHASE. */
    if ($debug) {

      // Stop count timer.
      timer_stop($timer);

      // Calculate memory usage.
      $used_memory = memory_get_usage() - $memory_usage_start;

      // Get array with debug data.
      $memcache_storage_debug_output = &drupal_static('memcache_storage_debug_output');

      // Log entry about memcache action.
      $memcache_storage_debug_output[] = array(
        'action' => 'set',
        'timer' => timer_read($timer),
        'memory' => $used_memory,
        'bin' => $cache_bin,
        'key' => urldecode($cache_id),
        'result' => $result ? 'HIT' : 'MISS',
      );

      // Increase timer to get new time results.
      self::$debug_timer++;
    }
    /* END DEBUG. LAST PHASE. */

    return $result;
  }

  /**
   * Replace existing data in memcache pool.
   * Provide debug logging.
   *
   * @see http://www.php.net/manual/en/memcache.replace.php
   */
  public static function replace($cache_id, $data, $expire, $memcache, $cache_bin = '') {

    // Check the debug mode.
    $debug = variable_get('memcache_storage_debug', FALSE);

    /* START DEBUG. FIRST PHASE. */
    if ($debug) {
      // Start count time spent on setting data to memcache.
      $timer = $cache_bin . '_' . self::$debug_timer;
      timer_start($timer);
      // Get memory usage before set new value into cache.
      $memory_usage_start = memory_get_usage();
    }
    /* STOP DEBUG. END OF FIRST PHASE. */

    // Replace data if exists.
    $result = FALSE;
    if ($memcache instanceof Memcache) {
      // Get compression mode.
      $compressed = variable_get('memcache_storage_compress_data', 0);
      $result = $memcache->replace($cache_id, $data, $compressed, $expire);
    }
    elseif ($memcache instanceof Memcached) {
      $result = $memcache->replace($cache_id, $data, $expire);
    }

    /* START DEBUG. LAST PHASE. */
    if ($debug) {

      // Stop count timer.
      timer_stop($timer);

      // Calculate memory usage.
      $used_memory = memory_get_usage() - $memory_usage_start;

      // Get array with debug data.
      $memcache_storage_debug_output = &drupal_static('memcache_storage_debug_output');

      // Log entry about memcache action.
      $memcache_storage_debug_output[] = array(
        'action' => 'replace',
        'timer' => timer_read($timer),
        'memory' => $used_memory,
        'bin' => $cache_bin,
        'key' => urldecode($cache_id),
        'result' => $result ? 'HIT' : 'MISS',
      );

      // Increase timer to get new time results.
      self::$debug_timer++;
    }
    /* END DEBUG. LAST PHASE. */

    return $result;
  }

  /**
   * Delete value from memcache pool.
   * Provide debug logging.
   *
   * @see http://www.php.net/manual/en/memcache.delete.php
   */
  public static function delete($cache_id, $memcache, $cache_bin = '') {

    // Check the debug mode.
    $debug = variable_get('memcache_storage_debug', FALSE);

    /* START DEBUG. FIRST PHASE. */
    if ($debug) {
      // Start count time spent on setting data to memcache.
      $timer = $cache_bin . '_' . self::$debug_timer;
      timer_start($timer);
      // Get memory usage before set new value into cache.
      $memory_usage_start = memory_get_usage();
    }
    /* STOP DEBUG. END OF FIRST PHASE. */

    // Delete cached data from memcached.
    $result = $memcache->delete($cache_id);

    /* START DEBUG. LAST PHASE. */
    if ($debug) {

      // Stop count timer.
      timer_stop($timer);

      // Calculate memory usage.
      $used_memory = memory_get_usage() - $memory_usage_start;

      // Get array with debug data.
      $memcache_storage_debug_output = &drupal_static('memcache_storage_debug_output');

      // Log entry about memcache action.
      $memcache_storage_debug_output[] = array(
        'action' => 'delete',
        'timer' => timer_read($timer),
        'memory' => $used_memory,
        'bin' => $cache_bin,
        'key' => urldecode($cache_id),
        'result' => $result ? 'HIT' : 'MISS',
      );

      // Increase timer to get new time results.
      self::$debug_timer++;
    }
    /* END DEBUG. LAST PHASE. */
  }

  /**
   * Immediately invalidates all existing items.
   */
  static public function flushClusters() {

    // Get memcache/memcached object.
    $memcache = self::openConnection();

    // Do not continue if can't connect to memcached pool.
    if (!$memcache) {
      return FALSE;
    }

    $indexes = &drupal_static('memcache_storage_bin_indexes');

    if (empty($indexes)) {
      $cache = MemcacheStorageAPI::get('memcache_storage_bin_indexes', $memcache);
      if (isset($cache['memcache_storage_bin_indexes'])) {
        $indexes = $cache['memcache_storage_bin_indexes'];
      }
      else {
        $indexes = array();
      }
    }

    // Simply increase index for each bin.
    // This index is a part of memcache key, so cache will request for data with a new keys.
    foreach ($indexes as $cache_bin => &$bin_index) {
      ++$bin_index;
    }

    return MemcacheStorageAPI::set('memcache_storage_bin_indexes', $indexes, 0, $memcache);
  }

  /**
   * Get name of cluster where cache bin should be stored.
   * Config loads from settings.php.
   *
   * @param $bin
   *   Cache bin name. i.e. 'cache' or 'cache_bootstrap'.
   *
   * @return string
   *   Cluster name.
   */
  static protected function getCacheBinCluster($bin) {

    // Example:
    // $conf['memcache_bins'] = array(
    //   'cache' => 'default',
    //   'cache_page' => 'pages',
    // );
    $cluster_bins = variable_get('memcache_bins', array());

    // Get server name for current cache bin.
    $cluster = 'default';
    if (!empty($cluster_bins[$bin])) {
      $cluster = $cluster_bins[$bin];
    }

    return $cluster;
  }

  /**
   * Trigger error if something goes wrong.
   *
   * @param $message
   * @param array $variables
   */
  static protected function logError($message, $variables = array()) {

    // We can't use watchdog because this happens in a bootstrap phase
    // where watchdog is non-functional. Register a shutdown handler
    // instead so it gets recorded at the end of page load.
    register_shutdown_function('watchdog', 'memcache_storage', $message, $variables, WATCHDOG_ERROR);
  }

}
